#include "image.h"
#define STB_IMAGE_IMPLEMENTATION
#define STBI_ONLY_PNG
#include "stb_image.h"

image::image(resource_manager& rm, ini_category const* ini, std::string const& path, std::vector<uint8_t> const& data) :
	_alpha(false), _borrowed_pointer(false)
{
	int w, h, n;
	auto d = reinterpret_cast<color*>(stbi_load_from_memory(data.data(), data.size(), &w, &h, &n, 4));
	_data = new color[w * h];
	for(int i = 0; i < w * h; ++i) {
		if((d[i] & 0xff000000u) == 0)
			_alpha = true;
		_data[i] = d[i];
		_data[i] = (d[i] & 0xff000000u) |
			((d[i] & 0x00ff0000u) >> 16) |
			(d[i] & 0x0000ff00u) |
			((d[i] & 0x000000ffu) << 16);
	}
	stbi_image_free(d);
	_bounds = recti(vec2i::zero, vec2i(w, h));
}

image::image(vec2i size) :
	_borrowed_pointer(false),
	_data(new color[size.size()]),
	_bounds(vec2i::zero, size),
	_alpha(true)
{}

image::image(vec2i size, color* data) :
	_borrowed_pointer(true),
	_data(data),
	_bounds(vec2i::zero, size),
	_alpha(true)
{}

image::image(image const& other) :
	_borrowed_pointer(false),
	_data(new color[other.size().size()]),
	_bounds(other.bounds()),
	_alpha(other._alpha)
{
	std::memcpy(_data, other._data, size().size() * 4);
}

image::~image() {
	if(!_borrowed_pointer && _data)
		delete[] _data;
}

vec2i const& image::size() const {
	return _bounds.size;
}

recti const& image::bounds() const {
	return _bounds;
}

color* image::raw_pointer() const {
	return _data;
}

color image::get(vec2i pos) const {
	return _data[pos.x + pos.y * _bounds.size.x];
}

color image::get_safe(vec2i pos) const {
	if(pos.x < 0 || pos.y < 0 || pos.x >= _bounds.size.x || pos.y >= _bounds.size.y)
		return 0x00000000;
	return _data[pos.x + pos.y * _bounds.size.x];
}

void image::clear(color color) {
	std::fill(_data, _data + size().size(), color);
}



